#include "rigidbody.h"
#include "camera.h"
#include "debug.h"
#include "program.h"
#include "collision.h"

typedef struct {
    Collision hit;
    float warming;
    int expiry;
} Contact;

struct RigidBody {
    Transformable transformable;

    float mass;

    Vector linear_force;
    Vector linear_velocity;

    int is_static;

    List contacts;
};

RigidBody* rigidbody_make(Transformable transform) {
    RigidBody* self = malloc(sizeof(RigidBody));
    ASSERT_RETURN(self != NULL, NULL, "Failed to allocate space for rigidbody");
    *self = (RigidBody){
        .transformable = transform,
        .mass = 1.0f,
        .linear_velocity = ZeroVector,
        .is_static = 0,
        .contacts = list_from_type(Contact),
    };
    return self;
}

void rigidbody_destroy(RigidBody* self) {
    list_empty(&self->contacts);
    free(self);
}

void rigidbody_add_contact(RigidBody* self, Collision hit) {
    LOG_INFO("contact between rigidbody %p and %p detected", self->transformable.data, hit.other.data);
    // update an existing contact
    list_foreach(Contact, contact, &self->contacts) {
        if(contact->hit.other.data == hit.other.data) {
            ++contact->expiry;
            contact->hit = hit;
            contact->warming += delta_time();
            return;
        }
    }

    // create a new contact
    list_add(&self->contacts,
    &(Contact) {
        .expiry = 1,
        .hit = hit,
        .warming = delta_time()
    });

    Contact* c = list_at(&self->contacts, self->contacts.len - 1);
    LOG_INFO("added contact %p with exp %d", c->hit.other.data, c->expiry);
}

static inline
void _internal_rigidbody_collect_contacts(RigidBody* self) {
    for(size_t i = 0; i < self->contacts.len; ++i) {
        Contact* contact = list_at(&self->contacts, i);
        --(contact->expiry);
        if(contact->expiry <= 0) {
            list_erase(&self->contacts, i);
            i--;
        }
    }
}

static inline
void _internal_debug_draw_collision_edge(Vector left, Vector right, Vector normal) {
#if !NDEBUG
    Vector a = camera_world_to_pixel_point(&g_camera, left);
    Vector b = camera_world_to_pixel_point(&g_camera, right);
    Vector n = transform_direction(&g_camera.transform, normal);
    SDL_SetRenderDrawColor(g_renderer, 255, 255, 255, 255);
    SDL_RenderDrawLine(g_renderer, a.x, a.y, b.x, b.y);
    a = vlerpf(a, b, 0.5f);
    b = vaddf(a, vmulff(n, 100.f));
    SDL_SetRenderDrawColor(g_renderer, 255, 0, 0, 255);
    SDL_RenderDrawLine(g_renderer, a.x, a.y, b.x, b.y);
#endif
}

static  inline
Vector _internal_calculate_contact_force(Contact* contact) {
    Collision hit = contact->hit;
    const float warming = fminf(1.f, contact->warming);
    const float d = 1.f;
    const float k = 50.0 * warming;
    const float b = 1.f;

    const Vector damping = vmulff(hit.normal, k * d);
    const Vector bounce = vprojectf(vmulff(hit.normal, -b), hit.velocity);

    return vsubf(damping, bounce);
}

static inline
void _internal_rigidbody_solve_contact(RigidBody* self, Contact* contact, Vector* solve_forces) {
    Collision hit = contact->hit;
    if (vsqrmagnitudef(contact->hit.separation_force) < 0.001f*0.001f)
        return;
    Vector force = _internal_calculate_contact_force(contact);
    float dot = vdotf(vnormalizedf(*solve_forces), force);
    if (veqf(*solve_forces, ZeroVector) || dot * dot > vsqrmagnitudef(*solve_forces) || dot <= 0.0f) {
        #if !NDEBUG
            LOG_INFO("warming: %f", contact->warming);
            LOG_INFO("force: %f %f", force.x, force.y);
            LOG_INFO("dot: %f", dot);
            LOG_INFO("mag: %f", vmagnitudef(self->linear_force));
        #endif
        *solve_forces = vaddf(*solve_forces, force);
        rigidbody_add_impulse(self, force);
    }

    _internal_debug_draw_collision_edge(hit.edge_left, hit.edge_right, hit.normal);

    ASSERT_RETURN(!visnanf(force), , "Force contains NaN (1)");
}

void rigidbody_solve_contacts(RigidBody* self) {
    ASSERT_RETURN(!visnanf(self->linear_velocity),, "Velocity is NaN (0)");

    Vector solve_forces = ZeroVector;
    list_foreach(Contact, contact, &self->contacts) {
        _internal_rigidbody_solve_contact(self, contact, &solve_forces);
    }

    _internal_rigidbody_collect_contacts(self);

    ASSERT_RETURN(!visnanf(self->linear_velocity),, "Velocity is NaN (1)");
}

void rigidbody_apply_physics(RigidBody* self) {
    Vector position = transformable_get_position(self->transformable);
    Vector velocity = vmulff(self->linear_velocity, delta_time());

    if(vsqrmagnitudef(velocity) > powf(0.00001f, 2)) {
        transformable_set_position(self->transformable, vaddf(position, velocity));
    }

    self->linear_force = ZeroVector;
}

float rigidbody_get_mass(const RigidBody* self) {
    return self->mass;
}

void rigidbody_set_mass(RigidBody* self, float mass) {
    self->mass = mass;
}

void rigidbody_add_impulse(RigidBody* self, Vector force) {
    self->linear_force = vaddf(self->linear_force, force);
    self->linear_velocity = vaddf(self->linear_velocity, force);
}

void rigidbody_accelerate(RigidBody* self, Vector force) {
    rigidbody_add_impulse(self, vmulff(force, delta_time()));
}

int rigidbody_is_static(const RigidBody* self) {
    return self->is_static;
}

void rigidbody_set_static(RigidBody* self, int is_static) {
    self->is_static = is_static;
}

Vector rigidbody_get_velocity(const RigidBody* self) {
    return self->linear_velocity;
}

void rigidbody_set_velocity(RigidBody* self, Vector velocity) {
    self->linear_force = vaddf(self->linear_force, vsubf(velocity, self->linear_force));
    self->linear_velocity = velocity;
}
